import copy
import logging
import signal
import time

import boto.exception
import boto.sqs
import boto.sqs.jsonmessage

from .persist import PersistStrategy
from .retry import UnlimitedRetries
from .utils import LogCapture, GracefulExit


class Consumer(object):
    def __init__(self,
                 region,
                 queue_name,
                 retry_strategy=None,
                 persist_on=None,
                 persist_backend=None,
                 queue_must_exist=True,
                 message_class=boto.sqs.jsonmessage.JSONMessage,
                 process_body_only=True):
        """
        Create an SQS Consumer for the configured queue.

        A `retry_strategy` will be consulted when a message fails processing
        to determine if it should be deleted or requeued. If none is supplied,
        messages will be retried indefinitely.
        """
        self.region = region
        self.queue_name = queue_name
        self.retry_strategy = retry_strategy or UnlimitedRetries
        self.persist_strategy = PersistStrategy(persist_on)
        self.persist_backend = persist_backend
        self.queue_must_exist = queue_must_exist
        self.message_class = message_class
        self.process_body_only = process_body_only
        self.logger = logging.getLogger(__name__)

        self._shutting_down = False

    def connect(self):
        """
        Connect to AWS and return the required boto.sqs.queue.Queue
        """
        if self._shutting_down:
            return None

        sqs = boto.sqs.connect_to_region(self.region)
        queue = sqs.get_queue(self.queue_name)
        if queue:
            queue.set_message_class(self.message_class)
        elif self.queue_must_exist:
            raise Exception(
                'Consumer expected queue "{}" to exist, but it doesn\'t!'
                .format(self.queue_name)
            )

        return queue

    def shutdown(self, signum, frame):
        # Stop consume() from processing any more messages
        self._shutting_down = True
        self.logger.warning('Caught shutdown signal, waiting for consumer to '
                            'finish processing current message...')

    def poll(self, processor, period):
        """
        Polls the queue and then sleeps for a specified period before trying
        again.
        """
        with GracefulExit(signal.SIGTERM, self.shutdown):
            while True:
                self.consume(self.connect(), processor)
                time.sleep(period)

    def consume(self, queue, processor):
        """
        Process messages received on the configured queue using the supplied
        processor function.
        The processor should return True if the message is considered to be
        successfully handled.
        Returns when the queue is empty.
        """
        while queue is not None:
            try:
                msgs = queue.get_messages(num_messages=1,
                                          attributes='All',
                                          message_attributes='All')
            except boto.exception.BotoClientError:
                self.logger.critical(
                    'Error occured while attempting to get message',
                    exc_info=True)
                return

            if not msgs:
                return
            message = msgs[0]

            attempt = int(message.attributes.get('ApproximateReceiveCount', 0))

            root_logger = logging.getLogger()
            with LogCapture(root_logger) as captured:
                if self.process_body_only:
                    body = copy.deepcopy(message.get_body())
                else:
                    body = message
                success = processor(body)

                retry = self.retry_strategy.should_retry(success, attempt)

                persist = self.persist_strategy.should_persist(success, retry)
                if persist:
                    self.persist_backend.persist(queue, message)

                if success or not retry:
                    queue.delete_message(message)

            if success:
                self.logger.debug(
                    'MessageID={id} successfully processed '
                    'on attempt #{n}, deleting.\n'
                    'Captured logs from message processing:\n{log}'
                    .format(id=message.id, n=attempt, log=captured.output())
                )
            elif not retry:
                self.logger.critical(
                    'MessageID={id} is being deleting because {reason}\n'
                    'Captured logs from message processing attempt:\n{log}'
                    .format(reason=self.retry_strategy,
                            id=message.id,
                            log=captured.output())
                )
            else:
                self.logger.warn(
                    'MessageID={id} failed on processing attempt #{n}, '
                    'message will be retried at a later time.\n'
                    'Captured logs from message processing attempt:\n{log}'
                    .format(id=message.id, n=attempt, log=captured.output())
                )



import logging

from .consumer import Consumer
from .message import AbstractMessage
from .persist import PersistBackend
from .processor import Processor
from .resource import AbstractResource
from .resource_factory import ResourceFactory
from .retry import RetryStrategy, LimitedRetries


logging.getLogger(__name__).addHandler(logging.NullHandler())


__all__ = [
    'Consumer', 'Processor',
    'AbstractMessage', 'AbstractResource',
    'ResourceFactory',
    'RetryStrategy', 'LimitedRetries',
    'PersistBackend',
]



import logging
import os
import signal
import StringIO
import sys
import time


class LogCapture(object):
    def __init__(self, logger):
        self.logger = logger
        self.buffer = StringIO.StringIO()
        self.handler = logging.StreamHandler(self.buffer)
        self._output = None

    def output(self):
        return self._output

    def __enter__(self):
        self.logger.addHandler(self.handler)
        return self

    def __exit__(self, type, value, traceback):
        self.logger.removeHandler(self.handler)

        self._output = self.buffer.getvalue()
        self.buffer.close()


shutdown_timeout = int(os.environ.get('CONSUMERISM_SHUTDOWN_TIME', 5))


class GracefulExit(object):
    def __init__(self, signum, handler, timeout=shutdown_timeout):
        self.signum = signum
        self.handler = handler
        self.timeout = timeout
        self.original_handler = signal.getsignal(signum)

    def graceful_handler(self, handler):
        def handle_and_exit_after_timeout(signum, frame):
            handler(signum, frame)
            time.sleep(self.timeout)
            sys.exit(0)
        return handle_and_exit_after_timeout

    def __enter__(self):
        signal.signal(self.signum, self.graceful_handler(self.handler))

    def __exit__(self, type, value, traceback):
        if type == KeyboardInterrupt:
            self.graceful_handler(self.handler)(signal.SIGINT, None)
        signal.signal(self.signum, self.original_handler)



import logging


class Echo(object):
    """
    A debug resource which will echo it's payload to the log.
    """

    def __init__(self, name):
        self.name = name
        self.logger = logging.getLogger(__name__)

    def invoke(self, *args, **kwargs):
        self.logger.info(
            'ECHO[{}]: args={} kwargs={}'.format(self.name, args, kwargs)
        )
        return True


class Failure(object):
    """
    A debug resource which will always fail, and therefore re-queue.
    """
    def __init__(self, name):
        self.name = name
        self.logger = logging.getLogger(__name__)

    def invoke(self, *args, **kwargs):
        self.logger.info(
            'FAILURE[{}]: failing and re-queueing...'.format(self.name)
        )
        return False



DISABLED = 'disabled'
DEAD = 'dead'
SUCCESS = 'success'
ALL = {DEAD, SUCCESS}


class PersistStrategy(object):
    """
    Strategy to determine when a message should be persisted.

    Configured by configuring which event(s) should cause persistence.
    """
    def __init__(self, events):
        if events is None:
            events = [DISABLED]
        if isinstance(events, basestring):
            events = [events]
        if not isinstance(events, set):
            events = set(events)

        self.events = events

    def should_persist(self, succeeded, will_retry):
        decision = (
            (DISABLED not in self.events)
            and
            (
                (succeeded and SUCCESS in self.events) or
                (not succeeded and not will_retry and DEAD in self.events)
            )
        )
        return decision


class PersistBackend(object):  # pragma: nocover
    """
    Abstract base class for persistence strategies.
    """
    def persist(self, queue, message):
        raise NotImplementedError



class RetryStrategy(object):  # pragma: nocover
    """
    Base class for retry strategies.
    """
    def should_retry(self, succeeded, attempt):
        """
        Returns True if a message should be eventually retried, given the
        most recent result and number of attempts at processing the message.
        """
        raise NotImplementedError

    def __repr__(self):
        return 'Retry Strategy logic descriptions goes here'


class LimitedRetries(RetryStrategy):
    """
    A strategy to limit the number of time a message is processes.
    Set limit to -1 for unlimited retries.
    """
    def __init__(self, limit):
        self.limit = limit

    def should_retry(self, succeeded, attempt):
        if succeeded:
            return False
        elif self.limit < 0:
            return True
        else:
            return attempt < self.limit

    def __repr__(self):  # pragma: nocover
        if self.limit < 0:
            return 'retry strategy allows unlimited processing attempts'
        elif self.limit == 1:
            return 'strategy configured to never retry'
        else:
            return 'retry strategy limited to {} attempts'.format(self.limit)


UnlimitedRetries = LimitedRetries(-1)

NeverRetry = LimitedRetries(1)



import logging


class Processor(object):
    """
    Processes SQS messages.
    """

    def __init__(self, resource_factory, message_class):
        """
        Create a new processor.
        The resource factory should return objects which implement the
        AbstractResource interface.
        """
        self.resource_factory = resource_factory
        self.message_class = message_class
        self.logger = logging.getLogger(__name__)

    def process(self, body):
        """
        Dispatch the JSON message to the correct Resource handler and
        perform the requested action.
        """
        try:
            message = self.message_class(body)
            resource_name = message.resource_name()
            invoke_args, invoke_kwargs = message.invoke_args()
        except Exception as e:
            # Discard malformed messages
            self.logger.warn(
                'Discarding malformed message, missing field "{}", body={}'
                .format(e, body))
            return True

        try:
            resource = self.resource_factory.new(resource_name)
        except:
            self.logger.critical(
                'Could not create resource "{}", body={}'
                .format(resource_name, body),
                exc_info=True)
            return False

        try:
            result = resource.invoke(*invoke_args, **invoke_kwargs)
        except:
            self.logger.critical(
                'Encountered exception while invoking resource "{}"'
                .format(resource_name),
                exc_info=True
            )
            return False

        return result



class AbstractResource(object):  # pragma: nocover
    """
    Base class for a resource.
    Do not inherit, just implement the interface.
    """

    logger = None

    def invoke(self, *args, **kwargs):
        """
        Invokes the named action with the supplied parameters.
        """
        raise NotImplementedError



class AbstractMessage(object):  # pragma: no cover
    """
    Base class for processed messages.
    """

    def __init__(self, body_dict):
        self.message = body_dict

    def resource_name(self):
        """
        Returns the name of the resource which should act upon this message.
        """
        raise NotImplementedError

    def invoke_args(self):
        """
        Returns arguments which should be passed to the resource's invoke
        method.
        """
        raise NotImplementedError



from setuptools import setup, find_packages


setup(
    name='consumerism',
    version='0.3.10',
    description='Expert360 Python SQS consumer library',
    author='Expert360',
    author_email='info@expert360.com',
    license='MIT',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2.7',
    ],
    keywords='boto sqs consumer',
    url='https://github.com/expert360/consumerism',
    packages=find_packages(exclude=['tests']),
    install_requires=[
        'boto>=2.34.0',
    ],
)



import logging
import os
import uuid

import boto.s3
import boto.sqs
import boto.sqs.jsonmessage

from .persist import PersistBackend


class DeadLetterS3Backend(PersistBackend):
    """
    A mechanism to store and reload dead letter messages to and from S3.
    Implements PersistBackend interface.
    """
    def __init__(self, region, bucket_name, queue_name,
                 message_class=boto.sqs.jsonmessage.JSONMessage,
                 *args, **kwargs):
        self.region = region
        self.bucket = bucket_name
        self.queue_name = queue_name
        self.message_class = message_class
        self.logger = logging.getLogger(__name__)

    def get_bucket(self):
        c = boto.s3.connect_to_region(self.region)
        bucket = c.get_bucket(self.bucket)
        return bucket

    def get_queue(self):
        sqs = boto.sqs.connect_to_region(self.region)
        queue = sqs.get_queue(self.queue_name)
        queue.set_message_class(self.message_class)
        return queue

    def persist(self, queue, message):
        """
        Persist a `boto.sqs.message` to the configured S3 bucket.
        """
        bucket = self.get_bucket()
        key = bucket.new_key(os.path.join(queue.name, message.id))
        key.set_contents_from_string(message.encode(message.get_body()))
        self.logger.info('Message {} persisted to s3://{}/{}, body={}'
                         .format(message.id, bucket.name, key.key,
                                 message.get_body()))


class DeadLetter(object):
    def __init__(self, backend_builder, *args, **kwargs):
        self.backend = backend_builder()
        self.logger = logging.getLogger(__name__)

    def _find_message_id(self, *args):
        # Find the first thing in the arg list which looks like a UUID
        for arg in args:
            try:
                return str(uuid.UUID(arg))
            except:
                continue
        return None

    def invoke(self, *args, **kwargs):
        # We try to be as compatible as possible with the implementors
        # Processor Message Class.
        # Furthermore, we never want any of this resources actions to result in
        # a re-queue

        action = None
        message_id = None
        try:
            if 'replay' in args:
                action = 'replay'

            if action:
                message_id = self._find_message_id(*args)

            if message_id:
                getattr(self, action)(message_id)

        except:
            self.logger.error(
                'DeadLetter failed: action={} message_id={}'
                .format(action, message_id),
                exc_info=True
            )
        finally:
            return True

    def replay(self, message_id):
        bucket = self.backend.get_bucket()
        queue = self.backend.get_queue()

        key = bucket.new_key(os.path.join(queue.name, message_id))
        message = queue.new_message()
        message.set_body(message.decode(key.get_contents_as_string()))
        replayed_message = queue.write(message)

        self.logger.info(
            'Message replayed from s3://{}/{} with new MessageID={}'
            .format(bucket.name, key.key, replayed_message.id)
        )



import functools
import inspect
import logging


class ResourceFactory(object):
    def __init__(self):
        self.resource_map = {}
        self.logger = logging.getLogger(__name__)

    def register(self, name, cls, **kwargs):
        """
        Register the given class with the factory.
        Any kwargs specified will be matched against the classes constructor
        arguments and automatically injected.
        """
        argspec = inspect.getargspec(cls.__init__)
        ctr_args = {k: v for k, v in kwargs.iteritems() if k in argspec.args}

        # A faux constructor for the class with DI from the supplied kwargs
        def constructor(cls, ctr_args, *args, **kwargs):
            kwargs.update(ctr_args)
            return cls(*args, **kwargs)

        self.resource_map[name] = functools.partial(constructor, cls, ctr_args)
        self.logger.info('Registered resource "{}"'.format(name))

    def new(self, name):
        resource_cls = self.resource_map.get(name, NoopResource)
        return resource_cls(name=name)


class NoopResource(object):
    """
    A no-operation resource which will simply print the fact that
    """
    def __init__(self, name):
        self.name = name
        self.logger = logging.getLogger(__name__)

    def invoke(self, *args, **kwargs):
        self.logger.warn(
            'NoopResource invoked for "{}" with arguments: args={} kwargs={}'
            .format(self.name, args, kwargs)
        )
        return True
